'''
An API to remotely control a pf based fail2ban
'''

import re
from ipaddress import ip_address
from subprocess import CalledProcessError

from flask import Flask, current_app, jsonify, request
from flask_httpauth import HTTPBasicAuth  # type: ignore
from werkzeug.security import check_password_hash

from jail2ban.auth import get_users
from jail2ban.pfctl import pfctl_cfg_read, pfctl_cfg_write, pfctl_table_op


auth = HTTPBasicAuth()


PAT_PORT = r'^any(?:\s+port\s+{\w+(?:,\w+)*})?$'
PAT_PROT = r'^(?:tcp|udp)$'
PAT_NAME = r'^[\w\-]+$'


def untaint(pattern: str, string: str) -> str:
    '''
    untaint string (as perl does)
    '''
    match = re.match(pattern, string)
    if match:
        return match.string
    raise ValueError(f'"{string}" is tainted')


def create_app():
    '''
    Create the flask application
    '''
    app = Flask(__name__, instance_relative_config=True)

    # load the instance config, if it exists, when not testing
    app.config.from_pyfile('config.py', silent=True)

    @auth.verify_password
    def verify_password(username: str, password: str) -> str | None:
        users = get_users()
        current_app.logger.debug(users)
        current_app.logger.debug('Checking password of %s', username)
        if username in users and \
                check_password_hash(users.get(username), password):
            return username
        return None

    @app.route("/ping", methods=['GET'])
    @auth.login_required
    def ping():
        remote_user = auth.username()
        app.logger.info('Received ping for'
                        ' anchor f2b-jail/%s', remote_user)
        return jsonify({'anchor': f'f2b-jail/{remote_user}',
                        'operation': 'ping',
                        'result': 'pong'})

    @app.route("/flush/<name>", methods=['GET'])
    @auth.login_required
    def flush(name):
        remote_user = auth.username()
        name = untaint(PAT_NAME, name)
        app.logger.info('Flushing table f2b-%s'
                        ' in anchor f2b-jail/%s', name, remote_user)
        res = pfctl_table_op('f2b-jail/{remote_user}',
                             table='f2b-{name}',
                             operation='flush')
        return jsonify({'anchor': f'f2b-jail/{remote_user}',
                        'table': f'f2b-{name}',
                        'operation': 'flush',
                        'result': [x.decode('ascii') for x in res]})

    @app.route("/register", methods=['PUT', 'DELETE'])
    @auth.login_required
    def register():
        remote_user = auth.username()
        # port / name / protocol
        data = request.get_json()
        name = untaint(PAT_NAME, data['name'])
        protocol = untaint(PAT_PROT, data['protocol'])
        port = untaint(PAT_PORT, data['port'])
        cfg = pfctl_cfg_read(f'f2b-jail/{remote_user}')
        if not cfg:
            cfg = []
        if request.method == 'PUT':
            cfg.extend([
                bytes(f'table <f2b-{name}> persist counters', 'ascii'),
                bytes(f'block quick proto {protocol}'
                      f' from <f2b-{name}> to {port}', 'ascii')
                ])
            res = pfctl_cfg_write(f'f2b-jail/{remote_user}',
                                  b'\n'.join(cfg) + b'\n')
        else:  # 'DELETE':
            cfg = [cfg_line for cfg_line in cfg
                   if cfg_line.find(bytes(f'<f2b-{name}>', 'ascii')) == -1]
            res = pfctl_cfg_write(f'f2b-jail/{remote_user}',
                                  b'\n'.join(cfg) + b'\n')
            pfctl_table_op(f'f2b-jail/{remote_user}',
                           table=f'f2b-{name}',
                           operation='flush')
            pfctl_table_op(f'f2b-jail/{remote_user}',
                           table=f'f2b-{name}',
                           operation='kill')
        app.logger.info('pfctl -a f2b-jail/%s -f-', remote_user)
        return jsonify({'anchor': f'f2b-jail/{remote_user}',
                        'table': f'f2b-{name}',
                        'action': 'start' if request.method == 'PUT'
                        else 'stop',
                        'result': [x.decode('ascii') for x in res]})

    @app.route("/ban", methods=['PUT', 'DELETE'])
    @auth.login_required
    def ban():
        remote_user = auth.username()
        data = request.get_json()
        # name / ip
        name = untaint(PAT_NAME, data['name'])
        ip = ip_address(data['ip'])
        if request.method == 'PUT':
            app.logger.info('Add %s to f2b-%s'
                            ' in anchor f2b-jail/%s', ip, name, remote_user)
            res = pfctl_table_op(f'f2b-jail/{remote_user}',
                                 table=f'f2b-{name}',
                                 operation='add',
                                 value=str(ip))
        else:  # 'DELETE':
            app.logger.info('Remove %s from f2b-%s'
                            ' in anchor f2b-jail/%s', ip, name, remote_user)
            res = pfctl_table_op(f'f2b-jail/{remote_user}',
                                 table=f'f2b-{name}',
                                 operation='delete',
                                 value=str(ip))
        return jsonify({'anchor': f'f2b-jail/{remote_user}',
                        'table': f'f2b-{name}',
                        'operation': 'add' if request.method == 'PUT'
                        else 'delete',
                        'result': res})

    @app.errorhandler(ValueError)
    def permission_err(error):
        '''
        Show a json parsable error if the value is illegal
        '''
        app.logger.fatal(error)
        return jsonify({'error': str(error)}), 500

    @app.errorhandler(CalledProcessError)
    def subprocess_err(error):
        '''
        Show a json parsable error if the value is illegal
        '''
        app.logger.fatal(error)
        return jsonify({'error': str(error)}), 500

    @app.errorhandler(FileNotFoundError)
    def filenotfound_err(error):
        '''
        Show a json parsable error if the value is illegal
        '''
        app.logger.fatal(error)
        return jsonify({'error': str(error)}), 500

    @auth.error_handler
    def auth_error():
        app.logger.error('Access Denied')
        return jsonify({'error': 'Access Denied'}), 401

    return app
